/*
 * s_http.c
 *
 *	Ver 1.0 : 2017/06/08
 *		Copy from:
 *			http://x68000.q-e-d.net/~68user/net/c-http-1.html
 *
 */

/* $Id: http-client.c,v 1.6 2013/01/23 06:57:19 68user Exp $ */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/param.h>
#include <sys/uio.h>
#include <unistd.h>
#include <fcntl.h>

/*
 *
 */

#include "s_http.h"

/*
 *
 */

#define BUF_LEN (256*10)                      /* バッファのサイズ */

/*
 * s_parse_url
 */

unsigned short s_parse_url ( char *host, char *path, char *url ) {

  unsigned short port = 80;            /* 接続するポート番号 */
  char host_path[BUF_LEN];

  strcpy ( host, "localhost" );
  strcpy ( path, "/" );

  if ( strlen(url) > BUF_LEN-1 ){
	fprintf(stderr, "URL が長すぎます。\n");
	return 0;
  }

  /* http:// から始まる文字列で */
  /* sscanf が成功して */
  /* http:// の後に何か文字列が存在するなら */
  if ( strstr(url, "http://") &&
	   sscanf(url, "http://%s", host_path) &&
	   strcmp(url, "http://" ) ){

	char *p;

	p = strchr(host_path, '/');  /* ホストとパスの区切り "/" を調べる */

	if ( p != NULL ){
	  strcpy(path, p);        /* "/"以降の文字列を path にコピー */
	  *p = '\0';
	  strcpy(host, host_path); /* "/"より前の文字列を host にコピー */
	} else {                     /* "/"がないなら＝http://host という引数なら */
	  strcpy(host, host_path); /* 文字列全体を host にコピー */
	}

	p = strchr(host, ':');       /* ホスト名の部分に ":" が含まれていたら */

	if ( p != NULL ){
	  port = atoi(p+1);        /* ポート番号を取得 */
	  if ( port <= 0 ){        /* 数字でない (atoi が失敗) か、0 だったら */
		port = 80;           /* ポート番号は 80 に決め打ち */
	  }
	  *p = '\0';
	}
  } else {
	fprintf(stderr, "URL は http://host/path の形式で指定してください。\n");
	return 0;
  }

  return port;
}

/*
 *	s_send_get_command
 */

int s_send_get_command ( char *host, char *path, unsigned short port ) {

  char send_buf[BUF_LEN];              /* サーバに送る HTTP プロトコル用バッファ */
  int s;                               /* ソケットのためのファイルディスクリプタ */
  struct hostent *servhost;            /* ホスト名と IP アドレスを扱うための構造体 */
  struct sockaddr_in server;           /* ソケットを扱うための構造体 */
  struct servent *service;             /* サービス (http など) を扱うための構造体 */

  servhost = gethostbyname(host);

  if ( servhost == NULL ){
	fprintf(stderr, "[%s] から IP アドレスへの変換に失敗しました。\n", host);
	return -1;
  }

  bzero(&server, sizeof(server));            /* 構造体をゼロクリア */

  server.sin_family = AF_INET;

  /* IPアドレスを示す構造体をコピー */
  bcopy(servhost->h_addr, &server.sin_addr, servhost->h_length);

  if ( port != 0 ){                          /* 引数でポート番号が指定されていたら */
	server.sin_port = htons(port);
  } else {                                   /* そうでないなら getservbyname でポート番号を取得 */
	service = getservbyname("http", "tcp");
	if ( service != NULL ){                /* 成功したらポート番号をコピー */
	  server.sin_port = service->s_port;
	} else {                               /* 失敗したら 80 番に決め打ち */
	  server.sin_port = htons(80);
	}
  }
  /* ソケット生成 */
  if ( ( s = socket(AF_INET, SOCK_STREAM, 0) ) < 0 ){
	fprintf(stderr, "ソケットの生成に失敗しました。\n");
	return -1;
  }
                                /* サーバに接続 */
  if ( connect(s, (struct sockaddr *)&server, sizeof(server)) == -1 ){
	fprintf(stderr, "connect に失敗しました。\n");
	return -1;
  }

                                /* HTTP プロトコル生成 & サーバに送信 */
  sprintf(send_buf, "GET %s HTTP/1.0\r\n", path);
  write(s, send_buf, strlen(send_buf));

  sprintf(send_buf, "Host: %s:%d\r\n", host, port);
  write(s, send_buf, strlen(send_buf));

  sprintf(send_buf, "\r\n");
  write(s, send_buf, strlen(send_buf));

  return s;

}

/*
 *	s_send_get
 */

char *s_send_get ( char *host, char *path, unsigned short port ) {
  static char r_buf[BUF_LEN];				/* 返す文字列を保持するバッファ */
  int s;                               /* ソケットのためのファイルディスクリプタ */

  if ( ( s = s_send_get_command ( host, path,  port ) ) != -1 ) {
                                /* あとは受信して、表示するだけ */

	int rbp = 0;
	int mode = 0;

	while (1){
	  int read_size;
	  char buf[BUF_LEN];

	  read_size = read(s, buf, BUF_LEN);
	  if ( read_size > 0 && mode < 4 ){
		int i;
		for ( i = 0; i < read_size && mode < 4; i++ ) {
		  switch ( mode ) {
		  case 0:
			if ( buf[i] == '\n' ) {
			  mode = 1;
			}
			break;
		  case 1:
			if ( buf[i] == '\r' ) {
			  mode = 2;
			} else if ( buf[i] == '\n' ) {
			  mode = 3;
			} else {
			  mode = 0;
			}
			break;
		  case 2:
			if ( buf[i] == '\n' ) {
			  mode = 3;
			} else {
			  mode = 0;
			}
			break;
		  case 3:
			if ( buf[i] == '\n' || buf[i] == '\r' ) {
			  mode = 4;
			  r_buf[rbp] = '\0';
			} else {
			  r_buf[rbp++] = buf[i];
			}
			break;
		  case 4:
			break;
		  }
		}
		//	  write(1, buf, read_size);
	  } else {
		break;
	  }
	}

	/* 後始末 */
	close(s);

  } else {
	return NULL;
  }
  
  return r_buf;
}

/*
 *	s_http_get
 */

char *s_http_get ( char *host, char *path ) {

  return s_send_get ( host, path, 80 );
}

/*
 *	s_url_get
 */

char *s_url_get ( char *url ) {
  char host[BUF_LEN];    /* 接続するホスト名 */
  char path[BUF_LEN];            /* 要求するパス */
  unsigned short port;            /* 接続するポート番号 */

  port = s_parse_url ( host, path, url );

  return s_send_get ( host, path, port );
}

/*
 *	s_http_transfer
 */

int s_http_transfer ( char *host, char *path, unsigned short port, char *save ) {
  int s;                   /* ソケットのためのファイルディスクリプタ */

  if ( ( s = s_send_get_command ( host, path,  port ) ) != -1 ) {
	FILE *wp;			/* ファイル保存 */

	if ( ( wp = fopen( save, "w" ) ) != NULL ) {
	  int mode = 0;

	  while (1){
		int read_size;
		char buf[BUF_LEN];

		read_size = read(s, buf, BUF_LEN);

		if ( read_size > 0 ) {
		  if ( mode < 4 ) {
			int i;
			for ( i = 0; i < read_size && mode < 4; i++ ) {
			  switch ( mode ) {
			  case 0:
				if ( buf[i] == '\n' ) {
				  mode = 1;
				}
				break;
			  case 1:
				if ( buf[i] == '\r' ) {
				  mode = 2;
				} else if ( buf[i] == '\n' ) {
				  mode = 3;
				} else {
				  mode = 0;
				}
				break;
			  case 2:
				if ( buf[i] == '\n' ) {
				  mode = 3;
				} else {
				  mode = 0;
				}
				break;
			  case 3:
				mode = 4;
				if ( i < read_size ) {
				  fwrite ( buf + i, read_size - i, 1, wp );
				}
				break;
			  }
			}
		  } else {
			fwrite ( buf, read_size, 1, wp );
		  }
		  //	  write(1, buf, read_size);
		} else {
		  break;
		}
	  }
	  fclose ( wp );
	} else {
	  fprintf ( stderr, "%s が Open できません\n", save );
	}
                                /* 後始末 */
	close(s);
  }

  return s;
}

/*
 *	s_http_download
 */

int s_http_download  ( char *host, char *path, char *save ) {

  return s_http_transfer ( host, path, 80, save );

}

/*
 *
 */

int s_url_download  ( char *url, char *save ) {
  char host[BUF_LEN];    /* 接続するホスト名 */
  char path[BUF_LEN];            /* 要求するパス */
  unsigned short port;            /* 接続するポート番号 */

  port = s_parse_url ( host, path, url );

  return s_http_transfer ( host, path, port, save );

}

#ifdef DEBUG

/*
 * main
 */

int main(int argc, char *argv[]){
  char host[BUF_LEN] = "localhost";    /* 接続するホスト名 */
  char path[BUF_LEN] = "/";            /* 要求するパス */
  unsigned short port = 80;            /* 接続するポート番号 */

  if ( argc > 1 ){                     /* URLが指定されていたら */
	port = s_parse_url ( host, path, argv[1] );
  }

  printf("http://%s%s を取得します。\n\n", host, path);

  if ( argc > 2 ) {
	s_http_transfer ( host, path, port, argv[2] );
  } else {
	printf ( "[%s]\n", s_send_get ( host, path, port ) );
  }

                                /* ホストの情報(IPアドレスなど)を取得 */
    return 0;
}

#endif
